from __future__ import annotations

import os
import tomllib
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


@dataclass(frozen=True)
class Config:
  workspace_path: Path = Path.home() / "workspace"
  additional_search_paths: list[Path] = field(default_factory=lambda: [])
  workspace_markers: list[str] = field(default_factory=lambda: ["*/.git"])
  hmrc: str = """new-session -d -c {{cwd}} -s {{session}} -n source nvim .
  new-window -c {{cwd}} -t {{session}} -n shell
  select-window -t {{session}}:source
  """

  @staticmethod
  def from_dict(data: dict[str, Any]) -> Config:  # pyright: ignore[reportExplicitAny]
    config_data = {}
    if "workspace_path" in data:
      config_data["workspace_path"] = Path(os.path.expanduser(data["workspace_path"]))  # pyright: ignore[reportAny]
    if "additional_search_paths" in data:
      config_data["additional_search_paths"] = [Path(os.path.expanduser(d)) for d in data["additional_search_paths"]]  # pyright: ignore[reportAny]
    if "workspace_markers" in data:
      config_data["workspace_markers"] = data["workspace_markers"]
    if "hmrc" in data:
      config_data["hmrc"] = data["hmrc"]
    return Config(**config_data)  # pyright: ignore[reportUnknownArgumentType]


def get_config_possible_paths() -> list[Path]:
  possible_paths = [
    (
      Path.home() / ".config"
      if "XDG_CONFIG_HOME" not in os.environ
      else Path(os.path.expanduser(os.environ["XDG_CONFIG_HOME"]))
    )
  ]
  if "XDG_CONFIG_DIRS" in os.environ:
    possible_paths.extend([Path(os.path.expanduser(p)) for p in os.environ["XDG_CONFIG_DIRS"].split(":")])
  return possible_paths


def read_config() -> Config:
  possible_paths = get_config_possible_paths()
  for possible_path in possible_paths:
    config_path = possible_path / "hivemux" / "config.toml"
    if not config_path.exists():
      continue
    with open(config_path, "rb") as file:
      read_config = tomllib.load(file)
      return Config.from_dict(read_config)
  # No match, so we return a default config file
  return Config()



from dataclasses import dataclass

from jinja2 import Template

from hivemux.config import Config
from hivemux.model import HivemuxProject, HivemuxSession
from hivemux.tmux import source_tmux_file


@dataclass(frozen=True)
class ProjectManager:
  config: Config

  def start_new_project(self, project: HivemuxProject) -> HivemuxSession:
    session_name = project.derive_session_name()
    possible_project_hmrc = project.path / ".hmrc"
    if possible_project_hmrc.exists():
      hmrc = possible_project_hmrc.read_text()
    else:
      hmrc = self.config.hmrc
    template = Template(hmrc)
    rendered_rc = template.render(session=session_name.to_tmux_session_name(), cwd=str(project.path.absolute()))  # pyright: ignore[reportAny]
    source_tmux_file(rendered_rc)
    return session_name



from dataclasses import dataclass
from pathlib import Path

from hivemux.model import HivemuxProject, HivemuxProjectName, HivemuxWorkspace


@dataclass(frozen=True)
class Workspace:
  human_friendly_name: str
  path: Path


def list_workspaces(
  base_workspace_path: HivemuxWorkspace, additional_search_paths: list[Path], workspace_markers: list[str]
) -> set[HivemuxProject]:
  projects: set[HivemuxProject] = set()
  for pattern in workspace_markers:
    for match in base_workspace_path.path.glob(pattern):
      parent_path = match.parent
      projects.add(HivemuxProject(path=parent_path, human_friendly_name=HivemuxProjectName(parent_path.name.lower())))
  for search_path in additional_search_paths:
    projects.add(HivemuxProject(path=search_path, human_friendly_name=HivemuxProjectName(search_path.name.lower())))
  return projects



import os
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile

from hivemux.model import HivemuxSession, HivemuxSessions


def new_session(cwd: Path, session_name: HivemuxSession, window_name: str, command: list[str] | None = None) -> None:
  session_cmd = " ".join(command) if command is not None else ""
  _ = subprocess.run(
    [
      "tmux",
      "new-session",
      "-d",
      "-c",
      cwd.absolute(),
      "-s",
      session_name.to_tmux_session_name(),
      "-n",
      window_name,
      session_cmd,
    ],
  )


def new_window(session_name: HivemuxSession, window_name: str) -> None:
  _ = subprocess.run(
    ["tmux", "new-window", "-t", session_name.to_tmux_session_name(), "-n", window_name],
  )


def has_session(session_name: HivemuxSession) -> bool:
  res = subprocess.run(["tmux", "has-session", "-t", session_name.to_tmux_session_name()])
  return res.returncode == 0


def activate_window(session_name: HivemuxSession, window_name: str) -> None:
  _ = subprocess.run(
    ["tmux", "select-window", "-t", f"{session_name.to_tmux_session_name()}:{window_name}"],
  )


def list_sessions() -> HivemuxSessions:
  res = subprocess.run(
    ["tmux", "list-sessions", "-F", "'#{session_name}'"],
    text=True,
    capture_output=True,
  )
  if res.returncode != 0:
    return HivemuxSessions([])
  sessions = [x.removeprefix("'").removesuffix("'") for x in res.stdout.splitlines()]
  return HivemuxSessions(
    [HivemuxSession.from_session_name(x) for x in sessions if HivemuxSession.is_hivemux_session(x)]
  )


def source_tmux_file(tmux_commands: str) -> None:
  with NamedTemporaryFile(
    "w",
    delete_on_close=False,
    delete=False,
  ) as f:
    _ = f.write(tmux_commands)
    f.flush()
    res = subprocess.run(["tmux", "source-file", str(Path(f.name).absolute())])
    res.check_returncode()


def switch_client(session_name: HivemuxSession) -> None:
  os.execvp("tmux", ["tmux", "switch-client", "-t", session_name.to_tmux_session_name()])


def attach(session_name: HivemuxSession) -> None:
  os.execvp("tmux", ["tmux", "attach", "-t", session_name.to_tmux_session_name()])



from pathlib import Path

import pytest

from hivemux.model import (
  CouldNotListWorkspacesException,
  HivemuxProject,
  HivemuxProjectName,
  HivemuxSession,
  HivemuxWorkspace,
  ProjectIsNotDirectoryException,
)


class TestHivemuxSession:
  @pytest.mark.parametrize(
    "session_name, tmux_session_name",
    [("a_name", "hm_a_name"), (".dot", "hm__dot"), ("a:project", "hm_a_project")],
  )
  def test_from_project_name_to_tmux_session_name(self, session_name: str, tmux_session_name: str) -> None:
    hivemux_session = HivemuxSession.from_project_name(HivemuxProjectName(session_name))

    assert hivemux_session.to_tmux_session_name() == tmux_session_name

  @pytest.mark.parametrize(
    "path, tmux_session_name",
    [
      ("/home/testuser/workspace/hello/", "hm_hello"),
      ("./workspace/.dot", "hm__dot"),
      ("/home/testuser/../othertestuser/:project", "hm__project"),
    ],
  )
  def test_from_path_to_tmux_session_name(self, path: str, tmux_session_name: str) -> None:
    hivemux_session = HivemuxSession.from_path(Path(path))

    assert hivemux_session.to_tmux_session_name() == tmux_session_name

  @pytest.mark.parametrize(
    "session_name, tmux_session_name",
    [("hm__dot", "hm__dot"), ("hm_project_name", "hm_project_name")],
  )
  def test_from_session_name_to_tmux_session_name(self, session_name: str, tmux_session_name: str) -> None:
    hivemux_session = HivemuxSession.from_session_name(session_name)

    assert hivemux_session.to_tmux_session_name() == tmux_session_name

  def test_from_session_fails_does_not_match_pattern(self) -> None:
    session_name_without_prefix = "hello"

    with pytest.raises(ValueError):
      _ = HivemuxSession.from_session_name(session_name_without_prefix)


class TestHivemuxWorkspace:
  def test_create_workspace_raises_with_non_directory(self) -> None:
    with pytest.raises(CouldNotListWorkspacesException):
      _ = HivemuxWorkspace(Path(__file__))

  def test_create_workspace_from_folder(self) -> None:
    path = Path(__file__).parent
    workspace = HivemuxWorkspace(path)
    assert workspace.path == path


class TestHivemuxProject:
  def test_project_does_not_point_to_folder_raises(self) -> None:
    with pytest.raises(ProjectIsNotDirectoryException):
      _ = HivemuxProject(Path(__file__), HivemuxProjectName("some_project"))

  def test_project_points_to_correct_folder(self) -> None:
    project = HivemuxProject(Path(__file__).parent, HivemuxProjectName("project_name"))
    assert project.derive_session_name() == HivemuxSession("tests")



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

SESSION_PREFIX = "hm_"

REPLACED_CHARS = [".", ":"]


def _cleanup_session_name(session_name: str) -> str:
  for char in REPLACED_CHARS:
    session_name = session_name.replace(char, "_")
  return session_name


@dataclass(frozen=True)
class HivemuxSession:
  session_name: str

  def to_tmux_session_name(self) -> str:
    return f"{SESSION_PREFIX}{self.session_name}"

  @staticmethod
  def is_hivemux_session(possible_session_name: str) -> bool:
    return possible_session_name.startswith(SESSION_PREFIX)

  @staticmethod
  def from_path(path: Path) -> HivemuxSession:
    return HivemuxSession(_cleanup_session_name(path.name.lower()))

  @staticmethod
  def from_session_name(value: str) -> HivemuxSession:
    if not HivemuxSession.is_hivemux_session(value):
      raise ValueError(f"Could not create HivemuxSession with non Hivemux session name {value}")
    return HivemuxSession(value.replace(SESSION_PREFIX, "").lower())

  @staticmethod
  def from_project_name(value: HivemuxProjectName) -> HivemuxSession:
    return HivemuxSession(_cleanup_session_name(value.value))


@dataclass(frozen=True)
class HivemuxSessions:
  sessions: list[HivemuxSession]

  def get_session_for_project(self, project: HivemuxProject) -> HivemuxSession | None:
    return next(
      iter([x for x in self.sessions if HivemuxSession.from_project_name(project.human_friendly_name) == x]),
      None,
    )


class CouldNotListWorkspacesException(Exception):
  def __init__(self, path: Path, reason: str) -> None:
    super().__init__(f"Could not list workspaces for {path.absolute()} because {reason}")


@dataclass(frozen=True)
class HivemuxWorkspace:
  path: Path

  def __post_init__(self) -> None:
    if not self.path.is_dir():
      raise CouldNotListWorkspacesException(self.path, "path is not a dir")


class ProjectIsNotDirectoryException(Exception):
  def __init__(self, path: Path) -> None:
    super().__init__(self, f"Project is not a directory {path}")


@dataclass(frozen=True)
class HivemuxProjectName(str):
  value: str


@dataclass(frozen=True)
class HivemuxProject:
  path: Path
  human_friendly_name: HivemuxProjectName

  def __post_init__(self) -> None:
    if not self.path.is_dir():
      raise ProjectIsNotDirectoryException(self.path)

  def derive_session_name(self) -> HivemuxSession:
    return HivemuxSession.from_path(self.path)



import os
import sys
from typing import override

import click
from click.shell_completion import CompletionItem

from hivemux import project as hivemux_project
from hivemux import tmux, workspaces
from hivemux import config
from hivemux.config import read_config
from hivemux.model import HivemuxProject, HivemuxWorkspace


@click.group()
def cli() -> None:
  pass


# @cli.command("list-sessions")
# def list_sessions() -> None:
#   sessions = tmux.list_sessions()
#   for session in sessions:
#     print(session)
#


def list_available_projects_from_config() -> set[HivemuxProject]:
  config = read_config()
  return workspaces.list_workspaces(
    HivemuxWorkspace(config.workspace_path),
    additional_search_paths=config.additional_search_paths,
    workspace_markers=config.workspace_markers,
  )


@cli.command("list-projects")
def list_projects() -> None:
  available_projects = list_available_projects_from_config()
  for workspace in sorted(available_projects, key=lambda x: x.human_friendly_name.lower()):
    print(workspace.human_friendly_name)


class ProjectVarType(click.ParamType):
  name = "project"

  @override
  def shell_complete(self, ctx: click.Context, param: click.Parameter, incomplete: str) -> list[CompletionItem]:
    return [
      CompletionItem(p.human_friendly_name)
      for p in list_available_projects_from_config()
      if p.human_friendly_name.startswith(incomplete)
    ]


@cli.command("attach")
@click.argument("project", type=ProjectVarType())
def attach(project: str) -> None:
  join_session(project)


@cli.command("a")
@click.argument("project", type=ProjectVarType())
def a(project: str) -> None:
  join_session(project)


def join_session(project: str) -> None:
  conf = config.read_config()
  project_manager = hivemux_project.ProjectManager(conf)
  available_projects = list_available_projects_from_config()
  matching_project = next(iter([p for p in available_projects if p.human_friendly_name == project.lower()]), None)
  if matching_project is None:
    print(f"Project {project} does not exist in workspace")
    sys.exit(1)
  current_tmux_sessions = tmux.list_sessions()
  possible_session = current_tmux_sessions.get_session_for_project(matching_project)
  if possible_session is None:
    possible_session = project_manager.start_new_project(matching_project)

  if "TMUX" in os.environ:
    tmux.switch_client(possible_session)
  else:
    tmux.attach(possible_session)
